@@ -1,3 +1,9 @@
+"""
+
+Neural network architectures
+
+"""
+
 import numpy as np
 import torch
 import torch.nn as nn
